@@ -1,10 +1,12 @@
-__version__ = "0.7.0"
+__version__ = "0.17.1"
 
 
 from .ast import *
 from .codegen import *
 from .helpers import *
 from .loop_info import *
+from .macro import *
+from .module import *
 from .parse import *
 from .plugin import *
 from .runtime import *
