@@ -8,8 +8,9 @@ __all__ = [
     "AstFormatString",
     "AstTuple",
     "AstList",
-    "AstDictItem",
     "AstDict",
+    "AstDictItem",
+    "AstDictUnquotedKey",
     "AstSlice",
     "AstUnpack",
     "AstKeyword",
@@ -25,19 +26,45 @@ __all__ = [
     "AstDecorator",
     "AstFunctionSignature",
     "AstFunctionSignatureArgument",
-    "AstFunctionRoot",
+    "AstDeferredRoot",
+    "AstMacro",
+    "AstMacroCall",
+    "AstMacroLiteral",
+    "AstMacroArgument",
+    "AstMacroMatch",
+    "AstMacroMatchLiteral",
+    "AstMacroMatchArgument",
+    "AstProcMacro",
+    "AstProcMacroMarker",
+    "AstProcMacroResult",
+    "AstClassName",
+    "AstClassBases",
+    "AstClassRoot",
     "AstInterpolation",
-    "AstImportedIdentifier",
+    "AstFromImport",
+    "AstImportedItem",
+    "AstImportedMacro",
 ]
 
 
+import re
 from dataclasses import dataclass
 from typing import Any, Optional, Union
 
-from beet.core.utils import required_field
-from mecha import AstChildren, AstNode, AstRoot
+from beet.core.utils import JsonDict, required_field
+from mecha import (
+    AstChildren,
+    AstCommand,
+    AstJson,
+    AstLiteral,
+    AstNode,
+    AstResourceLocation,
+    AstRoot,
+)
 from tokenstream import TokenStream
 
+from .pattern import IDENTIFIER_PATTERN
+
 
 @dataclass(frozen=True)
 class AstModuleRoot(AstRoot):
@@ -102,6 +129,11 @@ class AstList(AstExpression):
     items: AstChildren[AstExpression] = required_field()
 
 
+@dataclass(frozen=True)
+class AstDictUnquotedKey(AstValue):
+    """Ast dict unquoted key node."""
+
+
 @dataclass(frozen=True)
 class AstDictItem(AstNode):
     """Ast dict item node."""
@@ -238,12 +270,123 @@ class AstFunctionSignature(AstNode):
 
 
 @dataclass(frozen=True)
-class AstFunctionRoot(AstNode):
-    """Ast function root node."""
+class AstDeferredRoot(AstNode):
+    """Ast deferred root node."""
 
     stream: TokenStream = required_field()
 
 
+@dataclass(frozen=True)
+class AstMacro(AstCommand):
+    """Ast macro node."""
+
+    def get_command_tree(self) -> JsonDict:
+        tree_root: JsonDict = {"type": "root"}
+        tree: JsonDict = tree_root
+
+        for node in self.arguments:
+            if isinstance(node, AstMacroLiteral):
+                child = {"type": "literal"}
+                tree["children"] = {node.value: child}
+            elif isinstance(node, AstMacroMatchLiteral):
+                child = {"type": "literal"}
+                tree["children"] = {node.match.value: child}
+            elif isinstance(node, AstMacroMatchArgument):
+                child = {"type": "argument"}
+                child["parser"] = node.match_argument_parser.get_canonical_value()
+                if properties := node.match_argument_properties:
+                    child["properties"] = properties.evaluate()
+                tree["children"] = {node.match_identifier.value: child}
+            else:
+                break
+            tree = child
+
+        tree["executable"] = True
+
+        return tree_root
+
+
+@dataclass(frozen=True)
+class AstMacroCall(AstCommand):
+    """Ast macro call node."""
+
+
+@dataclass(frozen=True)
+class AstMacroLiteral(AstLiteral):
+    """Ast macro literal node."""
+
+    parser = "bolt_macro_literal"
+
+    regex = re.compile(r"[^#:\s()]+")
+
+
+@dataclass(frozen=True)
+class AstMacroArgument(AstLiteral):
+    """Ast macro argument node."""
+
+    parser = "bolt_macro_argument"
+
+    regex = re.compile(IDENTIFIER_PATTERN)
+
+
+@dataclass(frozen=True)
+class AstMacroMatch(AstNode):
+    """Base node for macro match."""
+
+
+@dataclass(frozen=True)
+class AstMacroMatchLiteral(AstMacroMatch):
+    """Ast macro match literal node."""
+
+    match: AstMacroLiteral = required_field()
+
+
+@dataclass(frozen=True)
+class AstMacroMatchArgument(AstMacroMatch):
+    """Ast macro match argument node."""
+
+    match_identifier: AstMacroArgument = required_field()
+    match_argument_parser: AstResourceLocation = required_field()
+    match_argument_properties: Optional[AstJson] = None
+
+
+@dataclass(frozen=True)
+class AstProcMacro(AstMacro):
+    """Ast proc macro node."""
+
+
+@dataclass(frozen=True)
+class AstProcMacroMarker(AstNode):
+    """Ast proc macro marker node."""
+
+
+@dataclass(frozen=True)
+class AstProcMacroResult(AstNode):
+    """Ast proc macro result node."""
+
+    commands: AstChildren[AstCommand] = required_field()
+
+
+@dataclass(frozen=True)
+class AstClassName(AstNode):
+    """Ast class name node."""
+
+    decorators: AstChildren[AstDecorator] = AstChildren()
+    value: str = required_field()
+
+
+@dataclass(frozen=True)
+class AstClassBases(AstNode):
+    """Ast class bases node."""
+
+    inherit: AstChildren[AstExpression] = AstChildren()
+
+
+@dataclass(frozen=True)
+class AstClassRoot(AstRoot):
+    """Ast class root node."""
+
+
 @dataclass(frozen=True)
 class AstInterpolation(AstNode):
     """Ast interpolation node."""
@@ -255,7 +398,23 @@ class AstInterpolation(AstNode):
 
 
 @dataclass(frozen=True)
-class AstImportedIdentifier(AstNode):
-    """Ast imported identifier node."""
+class AstFromImport(AstCommand):
+    """Ast from import node."""
 
-    value: str = required_field()
+    identifier: str = ""
+
+
+@dataclass(frozen=True)
+class AstImportedItem(AstNode):
+    """Ast imported item node."""
+
+    name: str = required_field()
+    identifier: bool = True
+
+
+@dataclass(frozen=True)
+class AstImportedMacro(AstNode):
+    """Ast imported macro node."""
+
+    name: str = required_field()
+    declaration: AstMacro = required_field()
