@@ -13,7 +13,7 @@ __all__ = [
 
 
 from contextlib import contextmanager
-from dataclasses import dataclass, field, fields
+from dataclasses import dataclass, field, fields, replace
 from typing import (
     Any,
     Dict,
@@ -24,11 +24,11 @@ from typing import (
     Optional,
     Tuple,
     Type,
+    Union,
     cast,
     overload,
 )
 
-from beet.core.utils import normalize_string
 from mecha import (
     AstChildren,
     AstCommand,
@@ -43,19 +43,28 @@ from .ast import (
     AstAssignment,
     AstAttribute,
     AstCall,
+    AstClassBases,
+    AstClassName,
     AstDict,
     AstDictItem,
     AstExpressionBinary,
     AstExpressionUnary,
     AstFormatString,
+    AstFromImport,
     AstFunctionSignature,
     AstIdentifier,
-    AstImportedIdentifier,
+    AstImportedItem,
+    AstImportedMacro,
     AstInterpolation,
     AstKeyword,
     AstList,
     AstLookup,
+    AstMacro,
+    AstMacroCall,
+    AstMacroMatchArgument,
+    AstProcMacro,
     AstSlice,
+    AstTarget,
     AstTargetAttribute,
     AstTargetIdentifier,
     AstTargetItem,
@@ -64,6 +73,7 @@ from .ast import (
     AstUnpack,
     AstValue,
 )
+from .module import CodegenResult, MacroLibrary
 
 
 @dataclass
@@ -72,9 +82,12 @@ class Accumulator:
 
     indentation: str = ""
     refs: List[Any] = field(default_factory=list)
+    macros: MacroLibrary = field(default_factory=dict)
+    macro_ids: Dict[str, int] = field(default_factory=dict)
     lines: List[str] = field(default_factory=list)
     counter: int = 0
     header: Dict[str, str] = field(default_factory=dict)
+    root_scope: bool = True
 
     last_condition: str = "_bolt_error_last_condition"
 
@@ -82,7 +95,7 @@ class Accumulator:
         """Return the source code."""
         header = "".join(
             f"{variable} = {expression}\n"
-            for expression, variable in self.header.items()
+            for variable, expression in self.header.items()
         )
 
         lines: List[str] = ["_bolt_lineno = "]
@@ -104,12 +117,12 @@ class Accumulator:
 
     def helper(self, name: str, *args: Any) -> str:
         """Emit helper."""
-        helper = f"_bolt_runtime.helpers[{name!r}]"
+        helper = f"_bolt_helper_{name}"
 
         if helper not in self.header:
-            self.header[helper] = f"_bolt_helper_{normalize_string(name)}"
+            self.header[helper] = f"_bolt_runtime.helpers[{name!r}]"
 
-        return f"{self.header[helper]}({', '.join(map(str, args))})"
+        return f"{helper}({', '.join(map(str, args))})"
 
     def replace(self, node: str, **kwargs: str) -> str:
         """Emit replace helper."""
@@ -117,8 +130,8 @@ class Accumulator:
 
     def missing(self) -> str:
         """Emit missing sentinel."""
-        self.header["_bolt_runtime.helpers['missing']"] = f"_bolt_helper_missing"
-        return f"_bolt_helper_missing"
+        self.header["_bolt_helper_missing"] = "_bolt_runtime.helpers['missing']"
+        return "_bolt_helper_missing"
 
     def children(self, nodes: Iterable[str]) -> str:
         """Emit children helper."""
@@ -151,6 +164,38 @@ class Accumulator:
         self.counter += 1
         return name
 
+    def make_macro(self, node: AstMacro) -> str:
+        """Add macro to macro library."""
+        macro = self.get_macro(node.identifier)
+        group = node.identifier.partition(":")[0]
+        if self.root_scope:
+            self.macros.setdefault(group, {})[macro, node] = None
+        return macro
+
+    def import_macro(self, resource_location: str, node: AstImportedMacro) -> str:
+        """Import macro into macro library."""
+        macro = self.get_macro(node.declaration.identifier)
+        group = node.declaration.identifier.partition(":")[0]
+        if self.root_scope:
+            self.macros.setdefault(group, {})[macro, node.declaration] = (
+                resource_location,
+                node.name,
+            )
+        return macro
+
+    def get_macro(self, name: str) -> str:
+        """Get macro."""
+        if name not in self.macro_ids:
+            self.macro_ids[name] = len(self.macro_ids)
+        return f"_bolt_macro{self.macro_ids[name]}"
+
+    def lineno(self, lineno: Any):
+        """Emit line number."""
+        if isinstance(lineno, AstNode) and not lineno.location.unknown:
+            lineno = lineno.location.lineno
+        if isinstance(lineno, int):
+            self.lines.append(f"!lineno {lineno}\n")
+
     @contextmanager
     def block(self):
         """Wrap statements in an indented block."""
@@ -163,12 +208,21 @@ class Accumulator:
 
     def statement(self, code: str, *, lineno: Any = None):
         """Emit statement."""
-        if isinstance(lineno, AstNode) and not lineno.location.unknown:
-            lineno = lineno.location.lineno
-        if isinstance(lineno, int):
-            self.lines.append(f"!lineno {lineno}\n")
+        self.lineno(lineno)
         self.lines.append(f"{self.indentation}{code}\n")
 
+    @contextmanager
+    def function(self, name: str, *args: str):
+        """Emit function."""
+        self.statement(f"def {name}({', '.join(args)}):")
+        with self.block():
+            previous_root = self.root_scope
+            self.root_scope = False
+            temp_counter = self.counter
+            yield
+            self.counter = temp_counter
+            self.root_scope = previous_root
+
     @contextmanager
     def if_statement(self, condition: str):
         """Emit if statement."""
@@ -222,7 +276,10 @@ class CommandCollector(ChildrenCollector):
 
     def add_dynamic(self, *args: str):
         for arg in args:
-            self.acc.statement(f"_bolt_runtime.commands.append({arg})")
+            if arg.startswith("*"):
+                self.acc.statement(f"_bolt_runtime.commands.extend({arg[1:]})")
+            else:
+                self.acc.statement(f"_bolt_runtime.commands.append({arg})")
 
 
 class RootCommandCollector(CommandCollector):
@@ -384,18 +441,27 @@ def visit_binding(
 class Codegen(Visitor):
     """Code generator."""
 
-    def __call__(self, node: AstRoot) -> Tuple[Optional[str], Optional[str], List[Any]]:  # type: ignore
+    def __call__(self, node: AstRoot) -> CodegenResult:  # type: ignore
         acc = Accumulator()
         result = self.invoke(node, acc)
+
         if result is None:
-            return None, None, acc.refs
-        elif len(result) != 1:
+            return CodegenResult()
+
+        if len(result) != 1:
             raise ValueError(
                 f"Expected single result for {node.__class__.__name__} {result!r}."
             )
+
         output = acc.make_variable()
         acc.statement(f"{output} = {result[0]}")
-        return acc.get_source(), output, acc.refs
+
+        return CodegenResult(
+            source=acc.get_source(),
+            output=output,
+            refs=acc.refs,
+            macros=acc.macros,
+        )
 
     @rule(AstNode)
     def fallback(
@@ -446,9 +512,7 @@ class Codegen(Visitor):
             for arg in signature.arguments
         ]
 
-        acc.statement(f"def {signature.name}({', '.join(arguments)}):")
-
-        with acc.block():
+        with acc.function(signature.name, *arguments):
             for arg in signature.arguments:
                 if arg.default:
                     acc.statement(f"if {arg.name} is {acc.missing()}:")
@@ -495,6 +559,105 @@ class Codegen(Visitor):
         acc.statement(statement)
         return []
 
+    @rule(AstMacro)
+    def macro(
+        self,
+        node: AstMacro,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        macro = acc.make_macro(
+            replace(node, arguments=AstChildren(node.arguments[:-1]))
+        )
+        arguments = [
+            arg.match_identifier.value
+            for arg in node.arguments
+            if isinstance(arg, AstMacroMatchArgument)
+        ]
+
+        with acc.function(macro, *arguments):
+            yield from visit_body(cast(AstRoot, node.arguments[-1]), acc)
+
+        return []
+
+    @rule(AstMacroCall)
+    def macro_call(
+        self,
+        node: AstMacroCall,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        macro = acc.get_macro(node.identifier)
+        result = yield from visit_generic(node, acc)
+        if result is None:
+            result = acc.make_ref(node)
+
+        macro_call = acc.helper("macro_call", "_bolt_runtime", macro, result)
+        result = acc.make_variable()
+        acc.statement(f"{result} = {macro_call}", lineno=node)
+
+        return [f"*{result}"]
+
+    @rule(AstProcMacro)
+    def proc_macro(
+        self,
+        node: AstProcMacro,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        macro = acc.make_macro(
+            replace(node, arguments=AstChildren(node.arguments[:-1]))
+        )
+
+        with acc.function(macro, "stream"):
+            yield from visit_body(cast(AstRoot, node.arguments[-1]), acc)
+
+        if acc.root_scope:
+            acc.header["_bolt_proc_macros"] = "{}"
+            acc.statement(f"_bolt_proc_macros[{node.identifier!r}] = {macro}")
+
+        return []
+
+    @rule(AstCommand, identifier="class:name:bases:body")
+    @rule(AstCommand, identifier="class:name:body")
+    def class_definition(
+        self,
+        node: AstCommand,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        name = cast(AstClassName, node.arguments[0])
+        body = cast(AstRoot, node.arguments[-1])
+
+        decorators: List[str] = []
+        for decorator in name.decorators:
+            value = yield from visit_single(decorator.expression, required=True)
+            decorators.append(value)
+
+        inherit: List[str] = []
+
+        if isinstance(bases := node.arguments[1], AstClassBases):
+            for base in bases.inherit:
+                result = yield from visit_single(base, required=True)
+                inherit.append(result)
+
+        joined_bases = f"({', '.join(inherit)})" if inherit else ""
+        acc.statement(f"class {name.value}{joined_bases}:", lineno=node)
+
+        with acc.block():
+            temp_start = acc.counter
+            yield from visit_body(body, acc)
+            temp_stop = acc.counter
+
+            for i in range(temp_start, temp_stop):
+                acc.statement(f"del _bolt_var{i}")
+
+            acc.counter = temp_start
+
+        for decorator, value in list(zip(name.decorators, decorators))[::-1]:
+            acc.statement(
+                f"{name.value} = {value}({name.value})",
+                lineno=decorator,
+            )
+
+        return []
+
     @rule(AstCommand, identifier="del:target")
     def del_statement(
         self,
@@ -533,9 +696,12 @@ class Codegen(Visitor):
         node: AstCommand,
         acc: Accumulator,
     ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
-        condition = yield from visit_single(node.arguments[0], required=True)
-        acc.statement(f"while {condition}:")
+        acc.statement(f"while True:")
         with acc.block():
+            condition = yield from visit_single(node.arguments[0], required=True)
+            acc.statement(f"if not {condition}:")
+            with acc.block():
+                acc.statement("break")
             yield from visit_body(cast(AstRoot, node.arguments[1]), acc)
         return []
 
@@ -570,9 +736,44 @@ class Codegen(Visitor):
         acc.statement("continue")
         return []
 
+    @rule(AstCommand, identifier="with:subcommand")
+    def with_statement(
+        self,
+        node: AstCommand,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        clauses: List[str] = []
+        bindings: List[Tuple[str, AstTarget]] = []
+
+        subcommand = cast(AstCommand, node.arguments[0])
+        body: Optional[AstNode] = None
+
+        while not body:
+            value = yield from visit_single(subcommand.arguments[0], required=True)
+
+            if isinstance(target := subcommand.arguments[1], AstTarget):
+                tmp = f"_bolt_with{len(bindings)}"
+                clauses.append(f"{value} as {tmp}")
+                bindings.append((tmp, target))
+            else:
+                clauses.append(value)
+
+            last_arg = subcommand.arguments[-1]
+            if isinstance(last_arg, AstRoot):
+                body = last_arg
+            elif isinstance(last_arg, AstCommand):
+                subcommand = last_arg
+
+        acc.statement(f"with {', '.join(clauses)}:", lineno=node)
+        with acc.block():
+            for tmp, target in bindings:
+                yield from visit_binding(target, "=", tmp, acc)
+            yield from visit_body(body, acc)
+
+        return []
+
     @rule(AstCommand, identifier="import:module")
     @rule(AstCommand, identifier="import:module:as:alias")
-    @rule(AstCommand, identifier="from:module:import:subcommand")
     def import_statement(
         self,
         node: AstCommand,
@@ -580,30 +781,8 @@ class Codegen(Visitor):
     ) -> Optional[List[str]]:
         module = cast(AstResourceLocation, node.arguments[0])
 
-        if node.identifier == "from:module:import:subcommand":
-            names: List[str] = []
-            subcommand = cast(AstCommand, node.arguments[1])
-
-            while True:
-                if isinstance(name := subcommand.arguments[0], AstImportedIdentifier):
-                    names.append(name.value)
-                if subcommand.identifier == "from:module:import:name:subcommand":
-                    subcommand = cast(AstCommand, subcommand.arguments[1])
-                else:
-                    break
-
-            if module.namespace:
-                acc.statement(
-                    f"{', '.join(names)} = _bolt_runtime.from_module_import({module.get_value()!r}, {', '.join(map(repr, names))})",
-                    lineno=node,
-                )
-            else:
-                for name in names:
-                    rhs = acc.get_attribute(acc.import_module(module.path), name)
-                    acc.statement(f"{name} = {rhs}", lineno=node)
-
-        elif node.identifier == "import:module:as:alias":
-            alias = cast(AstImportedIdentifier, node.arguments[1]).value
+        if node.identifier == "import:module:as:alias":
+            alias = cast(AstImportedItem, node.arguments[1]).name
 
             if module.namespace:
                 acc.statement(
@@ -625,6 +804,37 @@ class Codegen(Visitor):
 
         return []
 
+    @rule(AstFromImport)
+    def from_import_statement(
+        self,
+        node: AstFromImport,
+        acc: Accumulator,
+    ) -> Optional[List[str]]:
+        module = cast(AstResourceLocation, node.arguments[0])
+        items = cast(
+            AstChildren[Union[AstImportedItem, AstImportedMacro]], node.arguments[1:]
+        )
+        names = [item.name for item in items]
+
+        if module.namespace:
+            full_path = module.get_value()
+            targets: List[str] = []
+            for item in items:
+                if isinstance(item, AstImportedMacro):
+                    targets.append(acc.import_macro(full_path, item))
+                else:
+                    targets.append(item.name)
+            stmt = f"{', '.join(targets)} = _bolt_runtime.from_module_import({full_path!r}, {', '.join(map(repr, names))})"
+            acc.statement(stmt, lineno=node)
+        else:
+            stmt = f"_bolt_from_import = {acc.import_module(module.path)}"
+            acc.statement(stmt, lineno=node)
+            for name in names:
+                rhs = acc.get_attribute("_bolt_from_import", name)
+                acc.statement(f"{name} = {rhs}", lineno=node)
+
+        return []
+
     @rule(AstCommand, identifier="global:subcommand")
     @rule(AstCommand, identifier="nonlocal:subcommand")
     def global_nonlocal_statement(
@@ -975,3 +1185,33 @@ class Codegen(Visitor):
     ) -> Optional[List[str]]:
         acc.statement("pass")
         return []
+
+    @rule(AstCommand, identifier="raise")
+    def raise_statement(
+        self,
+        node: AstCommand,
+        acc: Accumulator,
+    ) -> Optional[List[str]]:
+        acc.statement("raise")
+        return []
+
+    @rule(AstCommand, identifier="raise:exception")
+    def raise_exc_statement(
+        self,
+        node: AstCommand,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        result = yield from visit_single(node.arguments[0], required=True)
+        acc.statement(f"raise {result}")
+        return []
+
+    @rule(AstCommand, identifier="raise:exception:from:cause")
+    def raise_exc_from_statement(
+        self,
+        node: AstCommand,
+        acc: Accumulator,
+    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
+        result = yield from visit_single(node.arguments[0], required=True)
+        cause = yield from visit_single(node.arguments[1], required=True)
+        acc.statement(f"raise {result} from {cause}")
+        return []
