@@ -14,6 +14,7 @@ from mecha import (
     AstBool,
     AstChildren,
     AstColor,
+    AstCommand,
     AstCoordinate,
     AstGamemode,
     AstGreedy,
@@ -44,6 +45,7 @@ from mecha import (
 )
 from tokenstream import set_location
 
+from .macro import invoke_macro
 from .utils import internal
 
 
@@ -61,6 +63,7 @@ def get_bolt_helpers() -> Dict[str, Any]:
         "get_rebind": get_rebind,
         "get_attribute": get_attribute,
         "import_module": python_import_module,
+        "macro_call": macro_call,
         "interpolate_bool": converter(AstBool.from_value),
         "interpolate_numeric": converter(AstNumber.from_value),
         "interpolate_coordinate": converter(AstCoordinate.from_value),
@@ -180,12 +183,19 @@ def python_import_module(name: str):
         raise exc.with_traceback(tb)
 
 
+@internal
+def macro_call(runtime: Any, function: Any, command: AstCommand):
+    return invoke_macro(runtime, function, command.identifier, command.arguments)
+
+
 def converter(f: Callable[[Any], AstNode]) -> Callable[[Any, AstNode], AstNode]:
     internal(f)
 
     @internal
     @wraps(f)
     def wrapper(obj: Any, node: AstNode) -> AstNode:
+        if isinstance(obj, AstNode):
+            return set_location(obj, node)
         return set_location(f(obj), node)
 
     return wrapper
@@ -199,6 +209,8 @@ class JsonObjectConverter:
 
     @internal
     def __call__(self, obj: Any, node: AstNode) -> AstNode:
+        if isinstance(obj, AstNode):
+            return set_location(obj, node)
         if isinstance(node := self.json_converter(obj, node), AstJsonObject):
             return node
         raise ValueError(f"Invalid json object of type {type(obj)!r}.")
@@ -212,6 +224,8 @@ class NbtCompoundConverter:
 
     @internal
     def __call__(self, obj: Any, node: AstNode) -> AstNode:
+        if isinstance(obj, AstNode):
+            return set_location(obj, node)
         if isinstance(node := self.nbt_converter(obj, node), AstNbtCompound):
             return node
         raise ValueError(f"Invalid nbt compound of type {type(obj)!r}.")
@@ -226,6 +240,8 @@ class EntityConverter:
 
     @internal
     def __call__(self, obj: Any, node: AstNode) -> AstNode:
+        if isinstance(obj, AstNode):
+            return set_location(obj, node)
         if isinstance(obj, str):
             if obj.count("-") == 4:
                 return self.uuid_converter(obj, node)
